//
//  main.cpp
//  kmeans_basic
//
//  Created by Masashi Morimoto on 2016/03/26.
//  Copyright © 2016年 AIT. All rights reserved.
//

/*** k平均法による領域抽出のプログラム ***/
#include <iostream>
#include <opencv2/opencv.hpp>

#define FILENAME "/Users/foobar/Pictures/fruit_image.jpg" // 画像ファイル名
#define MAX_CLUSTERS (6)    //クラスタの数 => K個
#define DIST_THRESHOLD (1)  //移動距離の閾値
#define HIGH_VAL (255)      // 白画素の値

//クラスタの初期化
void initCluster (std::vector<cv::Vec3d> *cluster_pt, cv::Mat img) {
    cluster_pt->clear();
    for (int i=0; i < MAX_CLUSTERS; i++) {
        int x = rand() % img.cols;
        int y = rand() % img.rows;
        cluster_pt->push_back(img.at<cv::Vec3b>(y,x));
    }
    return;
}

//距離算出
double calcDist (cv::Vec3d val1, cv::Vec3d val2) {
    return (sqrt((val1[0]-val2[0])*(val1[0]-val2[0]) + (val1[1]-val2[1])*(val1[1]-val2[1]) + (val1[2]-val2[2])*(val1[2]-val2[2])));
}

//クラスタのコピー
void copyCluster (std::vector<cv::Vec3d> cluster_pt, std::vector<cv::Vec3d> *cluster_pt_prev) {
    cluster_pt_prev->clear();
    for (int i=0; i<MAX_CLUSTERS; i++) {
        cluster_pt_prev->push_back(cluster_pt[i]);
    }
    return;
}

//クラスタと各点の距離を計算し，最も近いクラスタに移動させる
void setClusterOfPtr (std::vector<cv::Vec3d> cluster_pt, cv::Mat img, cv::Mat cluster_num) {
    double min_dist = DBL_MAX; //最小距離用の変数 (初期値としてdoubleの最大値を代入しておく)
    double dist; // 各点との距離
    cv::Vec3d val; // 各点の値
    for (int y=0; y < img.rows; y++) {
        for (int x = 0; x < img.cols; x++) {
            val = img.at<cv::Vec3b>(y,x);
            min_dist = DBL_MAX; //初期値としてdoubleの最大値を代入しておく
            for (int i=0; i<MAX_CLUSTERS; i++) {
                dist = calcDist(val, cluster_pt[i]); //距離の計算
                if (min_dist > dist) { //最小距離との比較
                    min_dist = dist; //最小距離の更新
                    cluster_num.at<int16_t>(y,x) = i; //クラスタの更新
                }
            }
        }
    }
    return;
}

//クラスタの平均位置の計算
void calcClusterCenter (std::vector<cv::Vec3d> *cluster_pt, cv::Mat img, cv::Mat cluster_num) {
    int pt_num[MAX_CLUSTERS];           //クラスタの対応点数
    std::vector<cv::Vec3d> cluster_sum; // 加算用領域
    //初期化
    for (int i=0; i < MAX_CLUSTERS; i++) {
        cluster_sum.push_back(cv::Vec3d(0,0,0));
        pt_num[i] = 0;
    }
    //クラスタの合計の算出
    for (int y=0; y < img.rows; y++) {
        for (int x = 0; x < img.cols; x++) {
            int idx = cluster_num.at<int16_t>(y,x);
            cluster_sum[idx] += img.at<cv::Vec3b>(y,x);
            pt_num[idx]++;
        }
    }
    //平均位置の算出
    cluster_pt->clear();
    for (int i=0; i < MAX_CLUSTERS; i++) {
        if (pt_num[i] > 0) {
            cluster_pt->push_back(cluster_sum[i]/pt_num[i]);
        } else {
            cluster_pt->push_back(cv::Vec3d(0,0,0));
        }
    }
    return;
}

//クラスタ中心色による画像の生成
void calcClusteringImage (std::vector<cv::Vec3d> cluster_pt, cv::Mat img, cv::Mat cluster_num) {
    for (int y=0; y < img.rows; y++) {
        for (int x = 0; x < img.cols; x++) {
            int idx = cluster_num.at<int16_t>(y,x);
            img.at<cv::Vec3b>(y,x) = cluster_pt[idx];
        }
    }
    return;
}

//クラスタの印字
void printCluster (std::vector<cv::Vec3d> cluster_pt) {
    for (int i=0; i<MAX_CLUSTERS; i++) {
        std::cout << "cluster:" << i << " Val: " << cluster_pt[i] << std::endl;
    }
    return;
}

// 変化量判定
bool isFinish (std::vector<cv::Vec3d> cluster_pt, std::vector<cv::Vec3d> cluster_pt_prev) {
    double dist = 0;
    for (int i=0; i < MAX_CLUSTERS; i++) {
        dist += calcDist(cluster_pt[i], cluster_pt_prev[i]); //点の距離の和
    }
    if (dist < DIST_THRESHOLD) { //この条件を満たすなら終了
        return (true);
    }
    return (false);
}

int main(int argc, const char * argv[])
{
    // 変数定義
    cv::Mat src_img, dst_img, label_img;
    cv::Vec3b val;
    std::vector<cv::Vec3d> cluster_pt; //クラスタ (中心)
    std::vector<cv::Vec3d> cluster_pt_prev; //クラスタ (中心，過去)

    // 画像入力
    src_img = cv::imread(FILENAME, cv::IMREAD_COLOR);
    if (src_img.empty()) { // 画像ファイル読み込み失敗
        printf("Cannot read image file: %s\n", FILENAME);
        return (-1);
    }
    dst_img = cv::Mat::zeros(src_img.size(), CV_8UC3);  // 出力画像（初期値 0）
    cv::Mat cluster_num = cv::Mat::ones(src_img.size(), CV_16SC1)*(-1); //所属するクラスタの番号 (初期値-1)
    initCluster(&cluster_pt, src_img);     // 1. 初期値設定
    do {
        copyCluster(cluster_pt, &cluster_pt_prev); //クラスタのコピー
        setClusterOfPtr(cluster_pt, src_img, cluster_num); // 2. 距離によるクラスタ分類
        calcClusterCenter(&cluster_pt, src_img, cluster_num); // 3. クラスタ中心算出
    } while (!isFinish(cluster_pt, cluster_pt_prev)); // 4. クラスタ中心の移動がほぼなければ終了

    // 特定ラベルの画素領域だけ描画
    label_img = cv::Mat::zeros(src_img.size(), CV_8UC1);
    for (int y=0; y<cluster_num.rows; y++) {
        for (int x=0; x<cluster_num.cols; x++) {
            int currentLabel = cluster_num.at<uint16_t>(y,x);
            if (currentLabel == 4) {
                label_img.at<unsigned char>(y,x)=HIGH_VAL;
            }
        }
    }
    cv::imshow("extracted region", label_img);
    //キー入力待ち
    cv::waitKey(0);
    return (0);
}